# Copyright 2025 Codon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LangGraph integration helpers for Codon Workloads."""
from __future__ import annotations

import inspect
import json
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, Union

from codon_sdk.agents import CodonWorkload
from codon_sdk.agents.codon_workload import WorkloadRuntimeError

from .callbacks import LangGraphTelemetryCallback

try:  # pragma: no cover - we do not require langgraph at install time
    from langgraph.graph import StateGraph  # type: ignore
except Exception:  # pragma: no cover
    StateGraph = Any  # fallback for type checkers

JsonDict = Dict[str, Any]
RawNodeMap = Mapping[str, Any]
RawEdgeIterable = Iterable[Tuple[str, str]]


def _merge_runtime_configs(
    base: Optional[Mapping[str, Any]],
    override: Optional[Mapping[str, Any]],
) -> Dict[str, Any]:
    merged: Dict[str, Any] = {}
    callbacks: List[Any] = []

    for cfg in (base, override):
        if not cfg:
            continue
        for key, value in cfg.items():
            if key == "callbacks":
                if isinstance(value, (list, tuple)):
                    callbacks.extend(value)
                else:
                    callbacks.append(value)
            else:
                merged[key] = value

    callbacks.append(LangGraphTelemetryCallback())
    merged["callbacks"] = callbacks
    return merged


@dataclass(frozen=True)
class LangGraphAdapterResult:
    """Artifacts produced when adapting a LangGraph graph."""

    workload: CodonWorkload
    state_graph: Any
    compiled_graph: Any


@dataclass(frozen=True)
class NodeOverride:
    """Caller-provided metadata overrides for a LangGraph node."""

    role: Optional[str] = None
    callable: Optional[Callable[..., Any]] = None
    model_name: Optional[str] = None
    model_version: Optional[str] = None
    input_schema: Optional[str] = None
    output_schema: Optional[str] = None


class LangGraphWorkloadAdapter:
    """Factory helpers for building Codon workloads from LangGraph graphs."""

    @classmethod
    def from_langgraph(
        cls,
        graph: Any,
        *,
        name: str,
        version: str,
        description: Optional[str] = None,
        tags: Optional[Sequence[str]] = None,
        org_namespace: Optional[str] = None,
        node_overrides: Optional[Mapping[str, Any]] = None,
        entry_nodes: Optional[Sequence[str]] = None,
        max_reviews: Optional[int] = None,
        compile_kwargs: Optional[Mapping[str, Any]] = None,
        runtime_config: Optional[Mapping[str, Any]] = None,
        return_artifacts: bool = False,
    ) -> Union[CodonWorkload, LangGraphAdapterResult]:
        """Create a :class:`CodonWorkload` from a LangGraph ``StateGraph``.

        Parameters
        ----------
        graph
            A LangGraph ``StateGraph`` (preferred) or compatible object exposing
            ``nodes``/``edges``.
        compile_kwargs
            Optional keyword arguments forwarded to ``graph.compile(...)`` so
            you can attach checkpointers, memory stores, or other runtime extras.
        return_artifacts
            When ``True`` return a :class:`LangGraphAdapterResult` containing the
            workload, the original state graph, and the compiled graph.
        """

        compiled, raw_nodes, raw_edges = cls._normalise_graph(
            graph, compile_kwargs=compile_kwargs
        )
        overrides = cls._normalise_overrides(node_overrides)
        node_map = cls._coerce_node_map(raw_nodes)
        raw_edge_list = cls._coerce_edges(raw_edges)

        node_names = set(node_map.keys())
        valid_edges = []
        entry_from_virtual = set()
        for src, dst in raw_edge_list:
            if src not in node_names or dst not in node_names:
                if src not in node_names and dst in node_names:
                    entry_from_virtual.add(dst)
                continue
            valid_edges.append((src, dst))

        workload = CodonWorkload(
            name=name,
            version=version,
            description=description,
            tags=tags,
        )

        successors: Dict[str, Sequence[str]] = cls._build_successor_map(valid_edges)
        predecessors: Dict[str, Sequence[str]] = cls._build_predecessor_map(valid_edges)

        for node_name, runnable in node_map.items():
            override = overrides.get(node_name)
            role = cls._derive_role(node_name, runnable, override.role if override else None)
            model_name = override.model_name if override else None
            model_version = override.model_version if override else None
            nodespec_kwargs: Dict[str, Any] = {}
            if override and override.input_schema is not None:
                nodespec_kwargs["input_schema"] = override.input_schema
            if override and override.output_schema is not None:
                nodespec_kwargs["output_schema"] = override.output_schema

            instrumented_callable = cls._wrap_node(
                node_name=node_name,
                role=role,
                runnable=runnable,
                successors=tuple(successors.get(node_name, ())),
                nodespec_target=override.callable if override else None,
                model_name=model_name,
                model_version=model_version,
                nodespec_kwargs=nodespec_kwargs or None,
            )
            workload.add_node(
                instrumented_callable,
                name=node_name,
                role=role,
                org_namespace=org_namespace,
            )

        for edge in valid_edges:
            workload.add_edge(*edge)

        workload._predecessors.update({k: set(v) for k, v in predecessors.items()})
        workload._successors.update({k: set(v) for k, v in successors.items()})

        if entry_nodes is not None:
            workload._entry_nodes = list(entry_nodes)
        else:
            inferred = [node for node, preds in predecessors.items() if not preds]
            inferred = list({*inferred, *entry_from_virtual})
            workload._entry_nodes = inferred or list(node_map.keys())

        setattr(workload, "langgraph_state_graph", graph)
        setattr(workload, "langgraph_compiled_graph", compiled)
        setattr(workload, "langgraph_compile_kwargs", dict(compile_kwargs or {}))
        setattr(workload, "langgraph_runtime_config", dict(runtime_config or {}))

        if return_artifacts:
            return LangGraphAdapterResult(
                workload=workload,
                state_graph=graph,
                compiled_graph=compiled,
            )

        return workload

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------
    @staticmethod
    def _normalise_graph(
        graph: Any, *, compile_kwargs: Optional[Mapping[str, Any]] = None
    ) -> Tuple[Any, Any, Any]:
        """Return compiled graph plus raw node/edge structures."""

        raw_nodes, raw_edges = LangGraphWorkloadAdapter._extract_nodes_edges(graph)
        compiled = graph
        if hasattr(graph, "compile"):
            kwargs = dict(compile_kwargs or {})
            compiled = graph.compile(**kwargs)
        comp_nodes, comp_edges = LangGraphWorkloadAdapter._extract_nodes_edges(compiled)

        nodes = raw_nodes or comp_nodes
        edges = raw_edges or comp_edges

        if nodes is None or edges is None:
            raise ValueError(
                "Unable to extract nodes/edges from LangGraph graph. Pass the original StateGraph or ensure the compiled graph exposes config.nodes/config.edges."
            )

        return compiled, nodes, edges

    @staticmethod
    def _extract_nodes_edges(obj: Any) -> Tuple[Optional[Any], Optional[Any]]:
        nodes = None
        edges = None

        graph_attr = getattr(obj, "graph", None)
        if graph_attr is not None:
            nodes = nodes or getattr(graph_attr, "nodes", None)
            edges = edges or getattr(graph_attr, "edges", None)

        nodes = nodes or getattr(obj, "nodes", None)
        edges = edges or getattr(obj, "edges", None)

        config = getattr(obj, "config", None)
        if config is not None:
            nodes = nodes or getattr(config, "nodes", None)
            edges = edges or getattr(config, "edges", None)
            if nodes is None and isinstance(config, Mapping):
                nodes = config.get("nodes")
            if edges is None and isinstance(config, Mapping):
                edges = config.get("edges")

        if nodes is not None and callable(getattr(nodes, "items", None)):
            nodes = dict(nodes)

        return nodes, edges

    @staticmethod
    def _coerce_node_map(nodes: Any) -> Dict[str, Any]:
        if isinstance(nodes, Mapping):
            result: Dict[str, Any] = {}
            for name, data in nodes.items():
                result[name] = LangGraphWorkloadAdapter._select_runnable(name, data)
            return result

        result: Dict[str, Any] = {}
        for item in nodes:
            if isinstance(item, tuple) and len(item) >= 2:
                name = item[0]
                data = item[1]
                result[name] = LangGraphWorkloadAdapter._select_runnable(name, data)
            else:
                raise ValueError(f"Unrecognized LangGraph node entry: {item!r}")

        return result

    @staticmethod
    def _normalise_overrides(overrides: Optional[Mapping[str, Any]]) -> Dict[str, NodeOverride]:
        if not overrides:
            return {}

        result: Dict[str, NodeOverride] = {}
        for name, value in overrides.items():
            if isinstance(value, NodeOverride):
                result[name] = value
                continue
            if isinstance(value, Mapping):
                result[name] = NodeOverride(
                    role=value.get("role"),
                    callable=value.get("callable"),
                    model_name=value.get("model_name"),
                    model_version=value.get("model_version"),
                    input_schema=value.get("input_schema"),
                    output_schema=value.get("output_schema"),
                )
                continue
            raise TypeError(
                "node_overrides values must be NodeOverride instances or mapping objects"
            )

        return result

    @staticmethod
    def _select_runnable(name: str, data: Any) -> Any:
        candidates: list[Any] = []

        if callable(data) or hasattr(data, "ainvoke") or hasattr(data, "invoke"):
            return data

        if isinstance(data, Mapping):
            for key in ("callable", "node", "value", "runnable", "invoke", "ainvoke"):
                if key in data and data[key] is not None:
                    candidates.append(data[key])
        else:
            for attr in ("callable", "node", "value", "runnable", "wrapped", "inner", "invoke", "ainvoke"):
                if hasattr(data, attr):
                    candidate = getattr(data, attr)
                    if candidate is not None and candidate is not data:
                        candidates.append(candidate)

        for candidate in candidates:
            if candidate is None:
                continue
            if callable(candidate) or hasattr(candidate, "ainvoke") or hasattr(candidate, "invoke"):
                return candidate

        raise WorkloadRuntimeError(f"Node '{name}' is not callable")

    @staticmethod
    def _coerce_edges(edges: Any) -> Sequence[Tuple[str, str]]:
        result: list[Tuple[str, str]] = []

        for item in edges:
            source = target = None
            if isinstance(item, tuple):
                if len(item) >= 2:
                    source, target = item[0], item[1]
            else:
                source = getattr(item, "source", None) or getattr(item, "start", None)
                target = getattr(item, "target", None) or getattr(item, "end", None)
                if source is None and isinstance(item, Mapping):
                    source = item.get("source")
                    target = item.get("target")

            if source is None or target is None:
                raise ValueError(f"Cannot determine edge endpoints for entry: {item!r}")

            result.append((source, target))

        return result

    @staticmethod
    def _derive_role(
        node_name: str,
        runnable: Any,
        override_role: Optional[str],
    ) -> str:
        if override_role:
            return override_role

        metadata = getattr(runnable, "metadata", None)
        if isinstance(metadata, Mapping):
            role = metadata.get("role") or metadata.get("tag")
            if isinstance(role, str):
                return role

        if "_" in node_name:
            return node_name.split("_")[0]
        return node_name

    @classmethod
    def _wrap_node(
        cls,
        *,
        node_name: str,
        role: str,
        runnable: Any,
        successors: Sequence[str],
        nodespec_target: Optional[Callable[..., Any]] = None,
        model_name: Optional[str] = None,
        model_version: Optional[str] = None,
        nodespec_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Callable[..., Any]:
        from codon.instrumentation.langgraph import track_node

        runnable = cls._unwrap_runnable(runnable)

        decorator = track_node(
            node_name=node_name,
            role=role,
            model_name=model_name,
            model_version=model_version,
            introspection_target=nodespec_target,
            nodespec_kwargs=nodespec_kwargs,
        )

        async def invoke_callable(state: Any, config: Optional[Mapping[str, Any]]) -> Any:
            if hasattr(runnable, "ainvoke"):
                try:
                    if config:
                        return await runnable.ainvoke(state, config=config)
                    return await runnable.ainvoke(state)
                except TypeError:
                    return await runnable.ainvoke(state)
            if inspect.iscoroutinefunction(runnable):
                return await runnable(state)
            if hasattr(runnable, "invoke"):
                try:
                    if config:
                        result = runnable.invoke(state, config=config)
                    else:
                        result = runnable.invoke(state)
                except TypeError:
                    result = runnable.invoke(state)
                if inspect.isawaitable(result):
                    return await result
                return result
            if callable(runnable):
                result = runnable(state)
                if inspect.isawaitable(result):
                    return await result
                return result
            raise WorkloadRuntimeError(f"Node '{node_name}' is not callable")

        @decorator
        async def node_callable(message: Any, *, runtime, context):
            if isinstance(message, Mapping) and "state" in message:
                state = message["state"]
            else:
                state = message

            workload = getattr(runtime, "_workload", None)
            base_config = None
            if workload is not None:
                base_config = getattr(workload, "langgraph_runtime_config", None)
            invocation_config = context.get("langgraph_config") if isinstance(context, Mapping) else None
            config = _merge_runtime_configs(base_config, invocation_config)

            result = await invoke_callable(state, config)

            if isinstance(result, Mapping):
                next_state: JsonDict = {**state, **result}
            else:
                next_state = {"value": result}

            for target in successors:
                runtime.emit(target, {"state": next_state})

            return next_state

        return node_callable

    @staticmethod
    def _unwrap_runnable(runnable: Any) -> Any:
        """Attempt to peel wrappers to find the actual callable runnable."""

        seen: set[int] = set()
        current = runnable

        while True:
            if current is None:
                break

            identifier = id(current)
            if identifier in seen:
                break
            seen.add(identifier)

            if hasattr(current, "ainvoke") or hasattr(current, "invoke") or callable(current):
                return current

            candidate = None
            for attr in ("callable", "node", "value", "wrapped", "inner", "runnable"):
                if hasattr(current, attr):
                    candidate = getattr(current, attr)
                    if candidate is not current:
                        break

            if candidate is None and isinstance(current, Mapping):
                for key in ("callable", "node", "value", "runnable"):
                    if key in current:
                        candidate = current[key]
                        if candidate is not current:
                            break

            if candidate is None:
                break

            current = candidate

        return runnable

    @staticmethod
    def _build_successor_map(edges: Sequence[Tuple[str, str]]) -> Dict[str, Sequence[str]]:
        successors: Dict[str, list] = defaultdict(list)
        for src, dst in edges:
            successors[src].append(dst)
        return {k: tuple(v) for k, v in successors.items()}

    @staticmethod
    def _build_predecessor_map(edges: Sequence[Tuple[str, str]]) -> Dict[str, Sequence[str]]:
        predecessors: Dict[str, list] = defaultdict(list)
        for src, dst in edges:
            predecessors[dst].append(src)
        return {k: tuple(v) for k, v in predecessors.items()}



# Copyright 2025 Codon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LangChain callback handlers that enrich Codon telemetry."""
from __future__ import annotations

import dataclasses
from typing import Any, Mapping, Optional

try:  # pragma: no cover - optional dependency
    from langchain.callbacks.base import BaseCallbackHandler
except Exception:  # pragma: no cover - fallback when LangChain is absent

    class BaseCallbackHandler:  # type: ignore
        """Minimal stand-in to keep instrumentation optional."""

        pass

from . import current_invocation


def _coerce_mapping(value: Any) -> Optional[Mapping[str, Any]]:
    if value is None:
        return None
    if isinstance(value, Mapping):
        return value
    if dataclasses.is_dataclass(value):  # pragma: no cover - defensive fallbacks
        return dataclasses.asdict(value)
    for attr in ("to_dict", "dict", "model_dump"):
        method = getattr(value, attr, None)
        if callable(method):
            result = method()
            if isinstance(result, Mapping):
                return result
    if hasattr(value, "__dict__"):
        data = {
            key: getattr(value, key)
            for key in vars(value)
            if not key.startswith("_")
        }
        return data
    return None


def _first(*values: Any) -> Optional[Any]:
    for value in values:
        if value:
            return value
    return None


def _normalise_usage(payload: Mapping[str, Any]) -> tuple[dict[str, Any], Optional[int], Optional[int], Optional[int]]:
    usage = {}
    for key in ("token_usage", "usage", "token_counts"):
        candidate = payload.get(key)
        if isinstance(candidate, Mapping):
            usage = dict(candidate)
            break

    # Some providers nest counts under token_count
    token_count = payload.get("token_count")
    if isinstance(token_count, Mapping):
        usage = usage or dict(token_count)

    # Provider-specific fallbacks
    for k in (
        "prompt_tokens",
        "input_tokens",
        "prompt_token_count",
        "input_token_count",
        "promptTokenCount",
        "inputTokenCount",
    ):
        if k in payload and k not in usage:
            usage[k] = payload[k]
    for k in (
        "completion_tokens",
        "output_tokens",
        "completion_token_count",
        "output_token_count",
        "completionTokenCount",
        "outputTokenCount",
    ):
        if k in payload and k not in usage:
            usage[k] = payload[k]
    if "total_tokens" not in usage and "totalTokenCount" in payload:
        usage["total_tokens"] = payload["totalTokenCount"]

    prompt_tokens = _first(usage.get("prompt_tokens"), usage.get("input_tokens"))
    completion_tokens = _first(
        usage.get("completion_tokens"), usage.get("output_tokens")
    )
    total_tokens = usage.get("total_tokens")
    if total_tokens is None and prompt_tokens is not None and completion_tokens is not None:
        total_tokens = prompt_tokens + completion_tokens

    return usage, prompt_tokens, completion_tokens, total_tokens


class LangGraphTelemetryCallback(BaseCallbackHandler):
    """Captures model metadata and token usage from LangChain callbacks."""

    def on_llm_start(self, serialized: Mapping[str, Any], prompts: list[str], **kwargs: Any) -> None:
        invocation = current_invocation()
        if not invocation:
            return

        params = _coerce_mapping(kwargs.get("invocation_params")) or _coerce_mapping(
            serialized.get("kwargs") if isinstance(serialized, Mapping) else None
        )

        identifier, vendor = _extract_model_info(params or {})

        if isinstance(serialized, Mapping):
            meta = _coerce_mapping(serialized.get("id"))
            serial_identifier, serial_vendor = _extract_model_info(serialized)
            identifier = identifier or serial_identifier
            vendor = vendor or _first(serial_vendor, meta.get("provider") if meta else None, serialized.get("name"))

        invocation.set_model_info(
            vendor=str(vendor) if vendor else None,
            identifier=str(identifier) if identifier else None,
        )

    def on_llm_end(self, response: Any, **kwargs: Any) -> None:
        invocation = current_invocation()
        if not invocation:
            return

        llm_output = _coerce_mapping(getattr(response, "llm_output", None))
        if llm_output:
            self._capture_payload(invocation, llm_output)

        response_metadata = _coerce_mapping(getattr(response, "response_metadata", None))
        if response_metadata:
            self._capture_payload(invocation, response_metadata)

        usage_metadata = _coerce_mapping(getattr(response, "usage_metadata", None))
        if usage_metadata:
            self._capture_payload(invocation, usage_metadata)

        generations = getattr(response, "generations", None)
        if generations:
            for generation_list in generations:
                for generation in generation_list:
                    metadata = getattr(generation, "generation_info", None)
                    if isinstance(metadata, Mapping):
                        self._capture_payload(invocation, metadata)

                    message = getattr(generation, "message", None)
                    if message is not None:
                        self._capture_message(invocation, message)

    def _capture_message(self, invocation, message: Any) -> None:
        for attr in ("usage_metadata", "response_metadata", "metadata"):
            payload = getattr(message, attr, None)
            if isinstance(payload, Mapping):
                self._capture_payload(invocation, payload)

        additional = getattr(message, "additional_kwargs", None)
        if isinstance(additional, Mapping):
            for key in ("usage_metadata", "response_metadata", "usageMetadata"):
                data = additional.get(key)
                if isinstance(data, Mapping):
                    self._capture_payload(invocation, data)

    def _capture_payload(
        self,
        invocation,
        payload: Mapping[str, Any],
    ) -> None:
        usage, prompt_tokens, completion_tokens, total_tokens = _normalise_usage(payload)
        if usage:
            invocation.record_tokens(
                input_tokens=prompt_tokens,
                output_tokens=completion_tokens,
                total_tokens=total_tokens,
                token_usage=usage,
            )

        model_identifier, model_vendor = _extract_model_info(payload)
        invocation.set_model_info(
            vendor=str(model_vendor) if model_vendor else None,
            identifier=str(model_identifier) if model_identifier else None,
        )

        response_metadata = _coerce_mapping(payload.get("response_metadata")) or _coerce_mapping(
            payload.get("metadata")
        )
        if response_metadata:
            invocation.add_network_call(dict(response_metadata))


def _extract_model_info(payload: Mapping[str, Any]) -> tuple[Optional[Any], Optional[Any]]:
    identifiers = (
        payload.get("model"),
        payload.get("model_name"),
        payload.get("model_id"),
        payload.get("modelName"),
    )
    vendors = (
        payload.get("model_vendor"),
        payload.get("provider"),
        payload.get("vendor"),
        payload.get("api_type"),
    )
    identifier = _first(*identifiers)
    vendor = _first(*vendors)
    if not identifier:
        meta = payload.get("response_metadata")
        if isinstance(meta, Mapping):
            identifier = _first(
                meta.get("model"),
                meta.get("model_name"),
                meta.get("model_id"),
            )
            vendor = _first(vendor, meta.get("model_vendor"), meta.get("provider"))
    return identifier, vendor


__all__ = ["LangGraphTelemetryCallback"]



# Copyright 2025 Codon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import json
import os
import time
from abc import ABC, abstractmethod
from contextvars import ContextVar
from functools import wraps
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence

from opentelemetry import trace
from opentelemetry.trace import Status, StatusCode

from .attributes import LangGraphSpanAttributes
from codon_sdk.instrumentation.schemas.nodespec import NodeSpec, NodeSpecSpanAttributes
from codon_sdk.agents import Workload
from codon_sdk.instrumentation.schemas.telemetry.spans import CodonBaseSpanAttributes
from codon_sdk.instrumentation.telemetry import NodeTelemetryPayload
from codon_sdk.instrumentation import initialize_telemetry

__all__ = [
    "LangGraphWorkloadMixin",
    "initialize_telemetry",
    "track_node",
    "LangGraphWorkloadAdapter",
    "LangGraphAdapterResult",
    "NodeOverride",
    "current_invocation",
    "LangGraphTelemetryCallback",
]

ORG_NAMESPACE: str = os.getenv("ORG_NAMESPACE")
__framework__ = "langgraph"

_instrumented_nodes: List[NodeSpec] = []


_ACTIVE_INVOCATION: ContextVar[Optional[NodeTelemetryPayload]] = ContextVar(
    "codon_langgraph_active_invocation", default=None
)


class LangGraphWorkloadMixin(ABC):
    """Mixin contract for workloads built from LangGraph graphs.

    Concrete implementations should inherit from this mixin *and* a concrete
    ``Workload`` subclass to reuse instrumentation helpers exposed here.
    """

    @classmethod
    @abstractmethod
    def from_langgraph(
        cls,
        graph: Any,
        *,
        name: str,
        version: str,
        description: Optional[str] = None,
        tags: Optional[Sequence[str]] = None,
    ) -> Workload:
        """Translate a LangGraph graph into a concrete Codon workload."""


def current_invocation() -> Optional[NodeTelemetryPayload]:
    """Return the currently active node invocation telemetry (if any)."""

    return _ACTIVE_INVOCATION.get()


def _safe_repr(value: Any, *, max_length: int = 2048) -> str:
    try:
        rendered = repr(value)
    except Exception as exc:  # pragma: no cover - defensive path
        rendered = f"<unrepresentable {type(value).__name__}: {exc}>"
    if len(rendered) > max_length:
        return rendered[: max_length - 3] + "..."
    return rendered


def _initial_input_payload(
    args: Sequence[Any], kwargs: Mapping[str, Any]
) -> Optional[str]:
    if args:
        return _safe_repr(args[0])
    if "message" in kwargs:
        return _safe_repr(kwargs["message"])
    if kwargs:
        return _safe_repr(dict(kwargs))
    return None


def _coerce_context(value: Any) -> Dict[str, Any]:
    if isinstance(value, Mapping):
        return dict(value)
    return {}


def _apply_nodespec_attributes(span, nodespec: NodeSpec) -> None:
    span.set_attribute(NodeSpecSpanAttributes.ID.value, nodespec.id)
    span.set_attribute(NodeSpecSpanAttributes.Version.value, nodespec.spec_version)
    span.set_attribute(NodeSpecSpanAttributes.Name.value, nodespec.name)
    span.set_attribute(NodeSpecSpanAttributes.Role.value, nodespec.role)
    span.set_attribute(
        NodeSpecSpanAttributes.CallableSignature.value,
        nodespec.callable_signature,
    )
    span.set_attribute(
        NodeSpecSpanAttributes.InputSchema.value,
        nodespec.input_schema,
    )
    if nodespec.output_schema is not None:
        span.set_attribute(
            NodeSpecSpanAttributes.OutputSchema.value,
            nodespec.output_schema,
        )
    if nodespec.model_name:
        span.set_attribute(NodeSpecSpanAttributes.ModelName.value, nodespec.model_name)
    if nodespec.model_version:
        span.set_attribute(
            NodeSpecSpanAttributes.ModelVersion.value,
            nodespec.model_version,
        )


def _apply_workload_attributes(
    span,
    *,
    telemetry: NodeTelemetryPayload,
    runtime: Any,
    nodespec: NodeSpec,
    context: Mapping[str, Any],
) -> None:
    workload = getattr(runtime, "_workload", None)

    organization = (
        telemetry.organization_id
        or telemetry.org_namespace
        or context.get("organization_id")
        or context.get("org_namespace")
        or (workload.organization_id if workload else None)
        or nodespec.org_namespace
        or ORG_NAMESPACE
    )

    span.set_attribute(
        CodonBaseSpanAttributes.AgentFramework.value,
        __framework__,
    )
    if organization:
        span.set_attribute(CodonBaseSpanAttributes.OrganizationId.value, organization)
        span.set_attribute(CodonBaseSpanAttributes.OrgNamespace.value, organization)
        telemetry.organization_id = telemetry.organization_id or organization
        telemetry.org_namespace = telemetry.org_namespace or organization

    workload_id = telemetry.workload_id or context.get("workload_id") or (
        workload.agent_class_id if workload else None
    )
    logic_id = (
        telemetry.workload_logic_id
        or context.get("logic_id")
        or context.get("workload_logic_id")
        or (workload.logic_id if workload else None)
    )
    run_id = (
        telemetry.workload_run_id
        or context.get("workload_run_id")
        or context.get("run_id")
    )
    deployment_id = telemetry.deployment_id or context.get("deployment_id")

    if workload_id:
        span.set_attribute(CodonBaseSpanAttributes.WorkloadId.value, workload_id)
        telemetry.workload_id = workload_id
    if logic_id:
        span.set_attribute(CodonBaseSpanAttributes.WorkloadLogicId.value, logic_id)
        telemetry.workload_logic_id = logic_id
    if run_id:
        span.set_attribute(CodonBaseSpanAttributes.WorkloadRunId.value, run_id)
        telemetry.workload_run_id = run_id
    if deployment_id:
        span.set_attribute(CodonBaseSpanAttributes.DeploymentId.value, deployment_id)
        telemetry.deployment_id = deployment_id

    workload_name = (
        telemetry.workload_name
        or context.get("workload_name")
        or (workload.metadata.name if workload else None)
    )
    workload_version = (
        telemetry.workload_version
        or context.get("workload_version")
        or (workload.metadata.version if workload else None)
    )

    if workload_name:
        span.set_attribute(
            CodonBaseSpanAttributes.WorkloadName.value,
            workload_name,
        )
        telemetry.workload_name = workload_name
    if workload_version:
        span.set_attribute(
            CodonBaseSpanAttributes.WorkloadVersion.value,
            workload_version,
        )
        telemetry.workload_version = workload_version


def _finalise_span(span, telemetry: NodeTelemetryPayload) -> None:
    if telemetry.node_input is not None:
        span.set_attribute(
            CodonBaseSpanAttributes.NodeInput.value,
            telemetry.node_input,
        )
        span.set_attribute(
            LangGraphSpanAttributes.Inputs.value,
            telemetry.node_input,
        )

    if telemetry.duration_ms is not None:
        span.set_attribute(
            CodonBaseSpanAttributes.NodeLatencyMs.value,
            telemetry.duration_ms,
        )
        span.set_attribute(
            LangGraphSpanAttributes.NodeLatency.value,
            f"{telemetry.duration_ms / 1000:.3f}",
        )

    if telemetry.node_output is not None:
        span.set_attribute(
            CodonBaseSpanAttributes.NodeOutput.value,
            telemetry.node_output,
        )
        span.set_attribute(
            LangGraphSpanAttributes.Outputs.value,
            telemetry.node_output,
        )

    span.set_attribute(
        CodonBaseSpanAttributes.NodeStatusCode.value,
        telemetry.status_code,
    )

    if telemetry.error_message:
        span.set_attribute(
            CodonBaseSpanAttributes.NodeErrorMessage.value,
            telemetry.error_message,
        )

    if telemetry.model_vendor:
        span.set_attribute(
            CodonBaseSpanAttributes.ModelVendor.value,
            telemetry.model_vendor,
        )

    if telemetry.model_identifier:
        span.set_attribute(
            CodonBaseSpanAttributes.ModelIdentifier.value,
            telemetry.model_identifier,
        )

    if telemetry.input_tokens is not None:
        span.set_attribute(
            CodonBaseSpanAttributes.TokenInput.value,
            telemetry.input_tokens,
        )

    if telemetry.output_tokens is not None:
        span.set_attribute(
            CodonBaseSpanAttributes.TokenOutput.value,
            telemetry.output_tokens,
        )

    if telemetry.total_tokens is not None:
        span.set_attribute(
            CodonBaseSpanAttributes.TokenTotal.value,
            telemetry.total_tokens,
        )

    if telemetry.token_usage:
        span.set_attribute(
            CodonBaseSpanAttributes.TokenUsageJson.value,
            json.dumps(telemetry.token_usage, default=str),
        )

    if telemetry.network_calls:
        span.set_attribute(
            CodonBaseSpanAttributes.NetworkCallsJson.value,
            json.dumps(telemetry.network_calls, default=str),
        )

    raw_json = telemetry.to_raw_attributes_json()
    if raw_json:
        span.set_attribute(
            CodonBaseSpanAttributes.NodeRawAttributes.value,
            raw_json,
        )


def track_node(
    node_name: str,
    role: str,
    model_name: Optional[str] = None,
    model_version: Optional[str] = None,
    introspection_target: Optional[Callable[..., Any]] = None,
    nodespec_kwargs: Optional[Mapping[str, Any]] = None,
):
    """Decorator to add telemetry instrumentation to a function.

    When the decorated function executes, this decorator:
    - Materializes a NodeSpec and captures its ID, signature, and schemas
    - Wraps execution in an OpenTelemetry span (async and sync supported)
    - Records inputs, outputs, and wall-clock latency via standardized span attributes

    Args:
        node_name: Unique identifier for this node.
        role: The node's role in the workflow.
        model_name: Optional model identifier if this node uses an AI model.
        model_version: Optional model version if this node uses an AI model.
        introspection_target: Optional callable to introspect instead of decorated function.
        nodespec_kwargs: Optional additional kwargs passed to NodeSpec constructor.

    Returns:
        The decorated function with telemetry instrumentation added.

    Example:
        >>> @track_node("retrieve_docs", role="retriever")
        ... def retrieve_docs(query: str) -> List[str]:
        ...     return ["doc1", "doc2"]

    TODO: Document what 'role' parameter specifically represents
    TODO: Clarify introspection_target use case and when to use it
    """
    def decorator(func):
        spec_callable = introspection_target or func
        spec_kwargs = dict(nodespec_kwargs or {})
        nodespec = NodeSpec(
            org_namespace=ORG_NAMESPACE,
            name=node_name,
            role=role,
            callable=spec_callable,
            model_name=model_name,
            model_version=model_version,
            **spec_kwargs,
        )
        _instrumented_nodes.append(nodespec)

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def awrapper(*args, **kwargs):
                tracer = trace.get_tracer(__name__)
                invocation_context = _coerce_context(kwargs.get("context"))
                runtime = kwargs.get("runtime")
                telemetry = getattr(runtime, "telemetry", None)
                if telemetry is None:
                    telemetry = NodeTelemetryPayload()
                if telemetry.node_input is None:
                    telemetry.node_input = _initial_input_payload(args, kwargs)
                telemetry.node_name = telemetry.node_name or nodespec.name
                telemetry.node_role = telemetry.node_role or nodespec.role
                telemetry.nodespec_id = telemetry.nodespec_id or nodespec.id
                token = _ACTIVE_INVOCATION.set(telemetry)

                try:
                    with tracer.start_as_current_span(nodespec.name) as span:
                        _apply_nodespec_attributes(span, nodespec)
                        _apply_workload_attributes(
                            span,
                            telemetry=telemetry,
                            runtime=runtime,
                            nodespec=nodespec,
                            context=invocation_context,
                        )

                        start = time.perf_counter()
                        try:
                            result = await func(*args, **kwargs)
                        except Exception as exc:  # pragma: no cover - surfacing telemetry
                            telemetry.status_code = "ERROR"
                            telemetry.error_message = repr(exc)
                            span.record_exception(exc)
                            span.set_status(Status(StatusCode.ERROR, str(exc)))
                            raise
                        else:
                            telemetry.node_output = _safe_repr(result)
                            return result
                        finally:
                            telemetry.duration_ms = int(
                                (time.perf_counter() - start) * 1000
                            )
                            _finalise_span(span, telemetry)
                finally:
                    _ACTIVE_INVOCATION.reset(token)

            return awrapper

        else:

            @wraps(func)
            def wrapper(*args, **kwargs):
                tracer = trace.get_tracer(__name__)
                invocation_context = _coerce_context(kwargs.get("context"))
                runtime = kwargs.get("runtime")
                telemetry = getattr(runtime, "telemetry", None)
                if telemetry is None:
                    telemetry = NodeTelemetryPayload()
                if telemetry.node_input is None:
                    telemetry.node_input = _initial_input_payload(args, kwargs)
                telemetry.node_name = telemetry.node_name or nodespec.name
                telemetry.node_role = telemetry.node_role or nodespec.role
                telemetry.nodespec_id = telemetry.nodespec_id or nodespec.id
                token = _ACTIVE_INVOCATION.set(telemetry)

                try:
                    with tracer.start_as_current_span(nodespec.name) as span:
                        _apply_nodespec_attributes(span, nodespec)
                        _apply_workload_attributes(
                            span,
                            telemetry=telemetry,
                            runtime=runtime,
                            nodespec=nodespec,
                            context=invocation_context,
                        )

                        start = time.perf_counter()
                        try:
                            result = func(*args, **kwargs)
                        except Exception as exc:  # pragma: no cover - surfacing telemetry
                            telemetry.status_code = "ERROR"
                            telemetry.error_message = repr(exc)
                            span.record_exception(exc)
                            span.set_status(Status(StatusCode.ERROR, str(exc)))
                            raise
                        else:
                            telemetry.node_output = _safe_repr(result)
                            return result
                        finally:
                            telemetry.duration_ms = int(
                                (time.perf_counter() - start) * 1000
                            )
                            _finalise_span(span, telemetry)
                finally:
                    _ACTIVE_INVOCATION.reset(token)

            return wrapper

    return decorator


from .adapter import (  # noqa: E402  # isort: skip
    LangGraphAdapterResult,
    LangGraphWorkloadAdapter,
    NodeOverride,
)
from .callbacks import LangGraphTelemetryCallback  # noqa: E402  # isort: skip
